#ifndef PEBBLISP_H
#define PEBBLISP_H

// #define STANDALONE

#define MAX_TOK_LEN 4
#define MAX_TOK_CNT 128 // 128
#define MAX_ENV_ELM 15 // 50

// static const char* tokenFail = "Missing ')'\n";

typedef struct Object Object;

struct Slice {
    const char *text;
    int length;
};

typedef enum Type {
    TYPE_NUMBER,
    TYPE_BOOL,
    TYPE_LIST,
    TYPE_FUNC,
    //TYPE_STRING,
    TYPE_SYMBOL,
    //TYPE_CONS,
    //TYPE_LAMBDA,
    //TYPE_MACRO,
    //TYPE_PRIMITIVE,
    TYPE_ENV,
    TYPE_ERROR
} Type;

struct Object {
    Type type;
    Object *forward;
    union {
        int number;
        Object *list;
        char name[MAX_TOK_LEN];
        Object (*func)(Object, Object);
        
        /*
        struct { double number; };                      // number
        struct { char string[sizeof (Object *[3])]; };  // string, symbol
        struct { Object *car, *cdr; };                  // cons
        struct { Object *params, *body, *env; };        // lambda, macro
        struct { int primitive; char *name; };          // primitive
        struct { Object *parent, *vars, *vals; };       // env
        struct { Object *forward; };                    // forwarding pointer
        */
    };
};

typedef struct Result {
    Object obj;
    struct Slice *slices;
} Result;

struct Environment {
    char **strings;
    Object *objects;
};

Object eval(const Object *obj, struct Environment *env);
char* stringObj(char *dest, const Object *obj);
Result parse(struct Slice *slices);
Result readSeq(struct Slice *slices);
Object parseAtom(struct Slice *slice);
void copySlice(char * dest, struct Slice *src);
Object parseEval(const char *input, struct Environment *env);
struct Environment defaultEnv();
void printObj(const Object *obj);

Result resultFromObjAndSlices(Object obj, struct Slice *slices);
Object add(Object obj1, Object obj2);

#define R(_obj, _slices) resultFromObjAndSlices(_obj, _slices)

#endif
